from dataclasses import dataclass
from itertools import groupby,chain
import inspect

import openpyxl
from openpyxl.styles import NamedStyle, PatternFill, Font, Alignment
from openpyxl.utils import get_column_letter

from datetime import datetime
from calendar import monthrange
from dateutil.relativedelta import relativedelta

from .marks import Marks
from .exceptions import DateNotFoundError
from .extra import get_student_year

def string_year(month_num : int | str , year : int | str) -> str :
  if 9 <= month_num <= 12:
    return year
  else :
    return year + 1
  


@dataclass
class _modify_journal_object:
  """
Модификцаия объекта Journal (Отсутствует параметр Subject)
  """
  mark : int
  date : datetime
  
  def __repr__(self) -> str:
    return f"{self.mark}"


@dataclass
class _subject_object:
  subject : str
  marks : list[_modify_journal_object]
  
  def __repr__(self) -> str:
    return f"{self.subject}{self.marks}"
  
  def average(self) -> float | int:
    marks = list(chain(*[i.mark for i in self.marks]))
    marks = [float(i) for i in marks if i.isdigit()]
    average = round(sum(marks) / len(marks),2)
    return f"{average:g}"
  
  def marks_row(self, days_quantity : int) -> list[str] :
    result = []
    only_dates = [int(datetime.strftime(i.date,'%d')) for i in self.marks]
    
    for day_num in range(1, days_quantity + 1):
      if day_num in only_dates:
        result.append(self.marks[only_dates.index(day_num)].mark)
      else:
        result.append('')
        
    return result


@dataclass
class _month_object:
  """
Объект месяца, содержащий информацию о каждом месяце , 
включая все оценки ,
которые были получены за этот месяц 
  """
  
  def __repr__(self) -> str:
    return f"{self.name().capitalize()} {self.data}"
  
  number : int
  quantity : int
  subjects_list : list
  data : list[_subject_object]
  
  def days_list(self) -> list[str] :
    days = [i for i in range(1,self.quantity +1)]
    while len(days) < 31:
      days.append(' ')
    return days
  
  def name(self) -> str:
    return ['январь', 'февраль', 'март', 'апрель', 'май', 'июнь', 'июль', 'август', 'сентябрь',
          'октябрь', 'ноябрь', 'декабрь'][self.number-1]
  

class _styles_object:
  
  def __init__(self, workbook : openpyxl.Workbook):
    self.wb = workbook
    
    # Автоматическое добавление стилей при иницализации объекта стилей
    # method_list = [method for method in dir(self.__class__) if method.startswith('__') is False]
    method_list = [m for m in self.__class__.__dict__.values() if inspect.isfunction(m) and m != __class__.__init__]
    for method in method_list:
      method(self)
  
  def neutral(self) -> None:
    ns = NamedStyle(name='neutral')
    ns.fill = PatternFill(fgColor='f3da0b', fill_type='solid')
    ns.font = Font(color='1f0800', name='Bahnscrift', bold=False)
    ns.alignment = Alignment(horizontal='center', vertical='center')
    self.wb.add_named_style(ns)
    
  def bad(self) -> None:
    ns = NamedStyle(name='bad')
    ns.fill = PatternFill(fgColor='ff8e7a', fill_type='solid')
    ns.font = Font(color='1f0800', name='Bahnscrift', bold=True)
    ns.alignment = Alignment(horizontal='center', vertical='center')
    self.wb.add_named_style(ns)
    
    


class ExcelTable:
  
  def __init__(self, marks : Marks, year : str | int) -> None:
    self.MARKS = marks.get_all()
    self.workbook = openpyxl.Workbook()
    self.sheet = self.workbook['Sheet']
    style = _styles_object(self.workbook)
    self.year = year
  
  def __month_generator(self):
    
    # Обрезка по году (от 20XX-20XX)
    croped_marks : list = [i for i in self.MARKS if i.date.strftime('%m.%Y') in get_student_year(self.year)]
    if not croped_marks:
      raise DateNotFoundError(f"Could not find any data for {self.year}")
    months : list = groupby(croped_marks, key = lambda x : x.date.strftime('%m'))
    

    for month, data in months:
      # Сортировка данных по уч.предметам
      data = list(data)
      data.sort(key = lambda x : x.subject)
      
      # Кол-во дней в месяце
      daysPM : int = monthrange(year = int(self.year),month=int(month))[1]
      
      # Группировка данных по уч.предметам
      lessonsDataPM = groupby(data, key=lambda y : y.subject)

      # Замена groupby объектов 
      lessonsModifyDataPM = []
      
      for subject_name , journals  in lessonsDataPM:
        journals = list(journals)
        journals = list(map(lambda x : _modify_journal_object(mark=x.mark,date=x.date),journals))
        lessonsModifyDataPM.append(_subject_object(subject = subject_name,
                                                   marks = list(journals)
                                                   ))

      # Список учебных предметов в месяце
      lessonsListPM : list = [i.subject for i in lessonsDataPM]
      
      MonthObject : _month_object =  _month_object(number = int(month),
                          data = lessonsModifyDataPM,
                          subjects_list= lessonsListPM,
                          quantity=daysPM)
      
      yield MonthObject
  
  def __recolor_bad_marks(self, bad_mark_indices : list[int], style : str) -> None : 
    for i in bad_mark_indices:
          self.sheet.cell(row = self.sheet.max_row,column=i).style = style
  
  def __fill(self) -> None:
    
    for month_object in self.__month_generator():
      header = [f"{month_object.name().capitalize()} {string_year(month_object.number,self.year)}"] + month_object.days_list() + ["Средний балл"]
      self.sheet.append(header)
      
      # Изменение шрифта для обозначения месяца
      self.sheet.cell(row = self.sheet.max_row, column=1).font = "Bahnschrift SemiBold"
      self.sheet.cell(row = self.sheet.max_row, column=1).alignment = Alignment(horizontal='center', vertical='center',wrap_text=True)
      
      for subject_object in month_object.data:
        body = [subject_object.subject] +  subject_object.marks_row(31) + [subject_object.average()]
        self.sheet.append(body)
        
        # Изменение шрифта и расположения для названия предмета 
        self.sheet.cell(row = self.sheet.max_row, column=1).font = "Bahnschrift SemiLight"
        
        
        
        # Выделение плохих оценок
        two_indices = [i+1 for i, x in enumerate(body[:-1]) if '2' in x]
        self.__recolor_bad_marks(two_indices,'bad')
        
        # Выделение средних оценок
        three_indices = [i+1 for i, x in enumerate(body[:-1]) if '3' in x]
        self.__recolor_bad_marks(three_indices,'neutral')
  
  def __post_processing(self) -> None:
    # Изменение размера ячеек
    for i in range(2, self.sheet.max_column+1):
      self.sheet.column_dimensions[f'{get_column_letter(i)}'].width = 5
    
    self.sheet.column_dimensions['AG'].width = 10
    self.sheet.column_dimensions['A'].width = 75
    
    # Изменение шрифтов 
    for row in self.sheet[f'B1:AO{self.sheet.max_row}']:
      for cell in row:
        cell.font = "Bahnschrift SemiLight"
        cell.alignment = Alignment(horizontal='center', vertical='center',wrap_text=True)

    
    # Изменение названия листа
    self.sheet.title = f'Оценки {self.year}-{self.year + 1}'
    
  def save(self, __file : str | bytes) -> None:
    # Заполнение листа данными
    self.__fill()
    # Пост-обработка
    self.__post_processing()
    
    self.workbook.save(__file)






from datetime import datetime
from dateutil.relativedelta import relativedelta

headers = {
    'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9',
    'Referer': 'https://cabinet.ruobr.ru/',
    'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.36',
}

def crop(__string: str, start:str, end:str) -> str:
  s = __string.find(start) + len(start)
  __string = __string[s:]
  e = __string.find(end)
  return __string[:e]

def get_student_year(__year : str) -> list :
  start = datetime.strptime(f'0109{__year}', "%d%m%Y").date()
  dates = [start.strftime('%m.%Y')]+ [(start + relativedelta(months=date)).strftime('%m.%Y') for date in range(1, 10)]
  return dates




import asyncio
import aiohttp
from bs4 import BeautifulSoup
from datetime import datetime
from dataclasses import dataclass
from itertools import groupby,chain
from .extra import crop, headers,get_student_year
from .exceptions import DateNotFoundError




@dataclass
class Journal():
  
  def __repr__(self) -> str:
    return f"{self.subject}[{self.mark}]"
  
  date : datetime
  subject : str 
  mark : str



def average(marks : list[Journal]) -> float | int:
  marks = list(chain(*[i.mark for i in marks]))
  marks = [float(i) for i in marks if i.isdigit()]
  average = round(sum(marks) / len(marks),2)
  return f"{average:g}"

class Marks():
  
  def __init__(self, __cookies : dict) -> None:
    self.__cookies = __cookies
  
  @staticmethod
  def __group_journals(journals : list[Journal]) -> list[Journal]:
    grouped_journals = {}
    for journal in journals:
        key = (journal.date, journal.subject)
        if key in grouped_journals:
            grouped_journals[key].mark += f", {journal.mark}"
        else:
            grouped_journals[key] = Journal(date=journal.date, mark=str(journal.mark), subject=journal.subject)
            
    # Преобразуем словарь обратно в список объектов Journal
    grouped_journals_list = list(grouped_journals.values())
    
    return grouped_journals_list
  
  def __marks(self) -> list[Journal]:

    async def pagination(session : aiohttp.ClientSession) -> int:
      response = await session.get('https://cabinet.ruobr.ru//student/progress/',cookies = self.__cookies,headers=headers)
      soup = BeautifulSoup(await response.text(), 'lxml')
      href_list = soup.find('ul',class_='pagination noprint').find_all('a')[-1]
      number = crop(str(href_list),'?page=', '"')
      return int(number)
    
    async def catch_page(url: str ,session : aiohttp.ClientSession):
      async with session.get(url=url,headers=headers,cookies = self.__cookies) as response:
        return await response.text()
        
    async def pages(timeout: float):
      tasks = []
      async with aiohttp.ClientSession() as session:
        pagination_num = await pagination(session)
        
        for num in range(pagination_num,0,-1):
          url = f"https://cabinet.ruobr.ru//student/progress/?page={num}"
          tasks.append(asyncio.create_task(catch_page(url,session))) 
          await asyncio.sleep(timeout)
          
        return await asyncio.gather(*tasks)
    
    async def sorter() -> list[Journal]:
      all_marks = []
      for page in await pages(0.0065):
        soup = BeautifulSoup(page, 'html.parser')
        data = list(map(str, soup.find_all('tr')))
        for t in data:
          cleantext = BeautifulSoup(t, 'html.parser').text.split('\n')[3:6]
          header = ['Дата', 'Дисциплина', 'Отметка']
          if cleantext != header:
            cleantext[2] = {'отлично': 5,'хорошо': 4,'удовлетворительно': 3,'неудовлетворительно': 2}[cleantext[2]]
            cleantext[0] = datetime.strptime(cleantext[0], '%d.%m.%Y')
            cleantext : Journal = Journal(date = cleantext[0], subject= cleantext[1],  mark = cleantext[2])
            all_marks.append(cleantext)
      
      
      all_marks.sort(key=lambda x: x.date)
      grouped_journals_list = self.__group_journals(all_marks)
      return grouped_journals_list
    
    async def main():
      return await sorter()
    
    return asyncio.run(main())

  def get_last(self) -> list[Journal]:
    import requests
    all_marks = []
    url = 'https://cabinet.ruobr.ru//student/progress/'
    response = requests.get(url,cookies=self.__cookies,headers=headers)
    soup = BeautifulSoup(response.text, 'html.parser')
    data = list(map(str, soup.find_all('tr')))
    for t in data:
      cleantext = BeautifulSoup(t, 'html.parser').text.split('\n')[3:6]
      header = ['Дата', 'Дисциплина', 'Отметка']
      if cleantext != header:
        cleantext[2] = {'отлично': 5,'хорошо': 4,'удовлетворительно': 3,'неудовлетворительно': 2}[cleantext[2]]
        cleantext[0] = datetime.strptime(cleantext[0], '%d.%m.%Y')
        cleantext : Journal = Journal(date = cleantext[0], subject= cleantext[1],  mark = cleantext[2])
        all_marks.append(cleantext)
            
    all_marks.sort(key=lambda x: x.date)
    grouped_journals_list = self.__group_journals(all_marks)
    return grouped_journals_list
    
  def get_all(self) -> list[Journal]:
    return self.__marks()

  def get_average(self, date : str) -> str:
    """
    param date - дата  для которой нужно вывести среднюю оценку по предметам
    """
    # Год
    if len(date.split('.')) == 1:
      croped_marks : list = [i for i in self.get_all() if i.date.strftime('%m.%Y') in get_student_year(date)]
    # Месяц
    elif len(date.split('.')) == 2:
      croped_marks : list = [i for i in self.get_all() if i.date.strftime('%m.%Y') in date]
    else: 
      raise DateNotFoundError(f"Could not find any data for {date}")
    
    if not croped_marks:
      raise DateNotFoundError(f"Could not find any data for {date}")
    
    
    croped_marks.sort(key=lambda x: x.subject)
    lessons = groupby(croped_marks, key=lambda y : y.subject)
    
    result = []
    for subject,marks in lessons:
      result.append(f'{subject}: {average(marks)}')
    return '\n'.join(result)
  
  def get_day_marks(self, date : str) -> str:
    croped_marks : list = [i for i in self.get_all() if i.date.strftime('%d.%m.%Y')==date]
    
    if not croped_marks:
      raise DateNotFoundError(f"Could not find any data for {date}")
    
    result = []
    for journal in croped_marks:
      result.append(f'{journal.subject}: {journal.mark}')
    return '\n'.join(result)
       
  


import requests
from datetime import datetime
from .extra import headers
from .exceptions import IncorrectDateError

class Schedule:

  def __init__(self, __cookies : dict)-> None:
    self.__cookies = __cookies
    
  def get_schedule(self, __date : str) -> list:
       
    correct_date = datetime.strptime(__date, "%d.%m.%Y").strftime('%Y-%m-%d')
    
    params = {'start': correct_date,'end': correct_date}
    response = requests.get('https://cabinet.ruobr.ru/student/get_schedule/g0/',
                            cookies=self.__cookies,
                            headers=headers,
                            params=params)
    response = response.json()
    
    if response:
      return response
    else :
      raise IncorrectDateError(f'Invalid date {__date}')

  def get_text(self, __date : str , formating : str = '|{time}| {lesson}') -> str:
     
    """
    Возвращает текст с расписанием по заданной дате\n
    :param __date: Дата для которой нужно получить расписание DD.MM.YYYY
    :param formating: Строка форматирования {time} - {lesson}
    :return: Отформатированный текст с расписанием
    """
    # Сортировка по времени
    data = sorted(self.get_schedule(__date), key= lambda x : x['start'])
    time,lesson  = [],[]
    
    for i in data:
      time.append(f"{i['start'].split('T')[1][:5]} - {i['end'].split('T')[1][:5]}")
      lesson.append(f'{i["title"].splitlines()[0]}')


    rows = []
    for time,lesson in zip(time,lesson):
      dk = {'time':time,'lesson':lesson}
      row = formating.format(**dk)
      rows.append(row)
    return '\n'.join(rows)  




from .marks import Marks
from .cookie import RuobrCookies
from .schedule import Schedule
from .excel import ExcelTable
from .exceptions import IncorrectDateError

class RuobrParser():
  
  def __init__(self,cookie : RuobrCookies) -> None:
    self.__cookies = cookie.cookies()
    
  def marks(self) -> Marks:
    return Marks(self.__cookies)

  def schedule(self):
    return Schedule(self.__cookies)
  
  def excel(self,year : int | str ) -> ExcelTable:
    if 3 < len(str(year)) == 4 and str(year).isdigit():
      return ExcelTable(self.marks(),int(year))
    else: 
      raise IncorrectDateError("Incorrect date format. Please enter a valid year in the format YYYY")


import requests
from .extra import headers
from .exceptions import AuthorizationError, RequestError

class RuobrCookies:
  
  """Класс для получения куки cabinet.ruobr.ru"""
  def __init__(self, username : str, password : str) -> None:
    self.username = username
    self.password = password
     
    cookies = self.cookies().values()
    if len(cookies) < 2:
      raise AuthorizationError(f"Invalid credantials in username or password")
    self.csrftoken,self.sessionid = self.cookies().values()

  def cookies(self) -> dict:
    """Cookie необходимые для авторизации"""
    data = {'username' : self.username, 'password' : self.password}
    session = requests.Session()
    
    response = session.get('https://cabinet.ruobr.ru/login/', headers=headers)
    if response.status_code != 200:
      raise RequestError(f"Status code [{response.status_code}]")
    data |= {'csrfmiddlewaretoken': dict(session.cookies)['csrftoken']}
    
    session.post('https://cabinet.ruobr.ru/login/', headers=headers, data=data)
    return dict(session.cookies)



class RuobrException(Exception):
  ...

class AuthorizationError(RuobrException):
  ...
  
class IncorrectDateError(RuobrException):
  ...

class EmptyDateError(RuobrException):
  ...

class RequestError(RuobrException):
  ...
  
class DateNotFoundError(RuobrException):
  ...